import bitalino

def connect_device(macAddress, client, devices):
    
    connected = False
    devices = [d for d in devices if d] # remove None
    print('devices: {}'.format(devices))
    
    if macAddress in [d.macAddress for d in devices]:
        try:
            print('{} state: {}'.format(macAddress, [d.state()for d in devices if d.macAddress==macAddress]))
            connected = True
        
        except Exception as e:
            print('error in connect_device: {}'.format(e))
            del devices[[d.macAddress for d in devices].index(macAddress)]
    
    else:

        try:
            device = bitalino.BITalino(macAddress, timeout=5)
            devices += [device]
            if macAddress in [d.macAddress for d in devices]:
                connected = True

            
        except Exception as e:
            print(e)

    devices = [d for d in devices if d] # remove None
    
    if not connected or macAddress not in [d.macAddress for d in devices]:
        client.publish(topic='rpi', qos=2, payload="['MAC STATE', '{}', '{}']".format(macAddress, 'failed'))
    else:
        client.publish(topic='rpi', qos=2, payload="['MAC STATE', '{}', '{}']".format(macAddress, 'connected'))
    
    return connected, devices


# built-in
from datetime import datetime
import os

# local
from epibox.bitalino.header2bitalino import header2bitalino
from epibox.scientisst.header2scientisst import header2scientisst


def open_file(directory, devices, mac_channels, sensors, fs, saveRaw, service):

    # for txt format
    now = datetime.now()
    save_time = now.strftime("%Y-%m-%d %H-%M-%S").rstrip('0')
    file_time = now.strftime("%Y-%m-%d %H:%M:%S.%f").rstrip('0')
    file_time = file_time[11:]
    file_time = '"' + file_time + '"'
    file_date = '"' + save_time[0:10] + '"'
    
    # we don't open the annot_file because it will be used in another thread
    annot_file = os.path.join(directory, 'ANNOT' + save_time +'.txt') #annotation file
    with open(annot_file, 'w') as fp:
        pass
    
    a_file = open(os.path.join(directory, 'A' + save_time + '.txt'), 'w') #data file
    
    if service == 'Bitalino' or service == 'Mini':
        save_fmt, header = header2bitalino(a_file, file_time, file_date, devices, mac_channels, sensors, fs, saveRaw, service)
    else:
        save_fmt, header = header2scientisst(a_file, file_time, file_date, devices, mac_channels, sensors, fs, saveRaw)
    
    drift_log_file = open(os.path.join(directory, 'drift_log_file_'+ save_time +'.txt'), 'w')


    return a_file, annot_file, drift_log_file, save_fmt, header



# built-in
import ast
import json
import random
import time
import string
import subprocess

# third-party
import paho.mqtt.client as mqtt

# local
from epibox.startup import startup 
from epibox.common.disconnect_system import disconnect_system
from epibox.common.create_folder import create_folder
from epibox.common.open_file import open_file
from epibox.common.write_file import write_annot_file
from epibox.common.start_system import start_system
from epibox.common.run_system import run_system
from epibox.common.connect_device import connect_device
from epibox.common import process_data


def random_str(length):
    letters = string.ascii_letters
    return ''.join(random.choice(letters) for i in range(length))

def on_message(client, userdata, message):
    
    message = str(message.payload.decode("utf-8"))
    message = ast.literal_eval(message)
    print("message received: ", message)
    
    if message[0] == 'RESTART':
        client.loop_stop()
        startup.main()
        
    elif message[0] == 'INTERRUPT':
        client.keepAlive = False
    
    elif message[0] == 'PAUSE ACQ':
        print('PAUSING ACQUISITION')
        global pause_acq
        pause_acq = True
        
    elif message[0] == 'RESUME ACQ':
        print('RESUMING ACQUISITION')
        pause_acq = False
        
    elif message[0] == 'ANNOTATION':
        print('RECEIVED ANNOT {} ----------------------'.format(message[1]))
        global new_annot
        global write_annot
        new_annot = message[1]
        write_annot = True
        
    elif message[0] == 'TURN OFF':
        print('TURNING OFF RPI')
        client.publish(topic='rpi', payload=str(['TURNED OFF']))
    
    elif message[0] == 'TURNED OFF':
        subprocess.run(['sudo', 'shutdown', '-h', 'now'])

#****************************** MAIN SCRIPT ***********************************

def main(devices):
    
    try:
        
        client_name = random_str(6)
        print('Client name (acquisition):', client_name)
        host_name = '192.168.0.10'
        topic = 'rpi'
        
        client = mqtt.Client(client_name)
        setattr(client, 'keepAlive', True)
        client.username_pw_set(username='preepiseizures', password='preepiseizures')
        client.connect(host_name)
        client.subscribe(topic)
        client.on_message = on_message
        client.loop_start()
        print('Successfully subcribed to topic', topic)
        init = False
        client.publish('rpi', "['STARTING']")

        with open('/home/pi/Documents/epibox/args.json', 'r') as json_file:
            opt = json_file.read()
            
        opt = ast.literal_eval(opt)
        
        if not opt['channels']:
            channels = []
            for device in opt['devices_mac']:
                for i in range(1,7):
                    channels += [[device,str(i)]]
            sensors = ['-' for i in range(len(channels))]
        
        else:
            channels = []
            sensors = []
            for triplet in opt['channels']:
                channels += [triplet[:2]] 
                sensors += [triplet[2]]
        
        saveRaw = bool(opt['saveRaw'])
        service = opt['service']
        
        global write_annot
        global new_annot
        global pause_acq

        write_annot = False
        pause_acq = False
        already_notified_pause = False
        system_started = False
        
        print('ID: {}'.format(opt['patient_id']))
        print('folder: {}'.format(opt['initial_dir']))
        print('fs: {}'.format(opt['fs']))
        print('saveRaw: {}'.format(saveRaw))
        print('channels: {}'.format(channels))
        print('devices: {}'.format(opt['devices_mac']))
        print('sensors: {}'.format(sensors))
        print('service: {}'.format(service))
        print('Devices in list: {}'.format([d.macAddress for d in devices]))
                            
                            
        # Use/create the patient folder =============================================================== 
        directory = create_folder(opt['initial_dir'], '{}'.format(opt['patient_id']), service)
        already_timed_out = False   
        
        try:
            a_file, annot_file, drift_log_file, save_fmt, header = open_file(directory, devices, channels, sensors, opt['fs'], saveRaw, service)

        except Exception as e:
            print(e)
            client.publish('rpi', "['ERROR']")
            client.loop_stop()
            
            # Disconnect the system
            print('Could not open the files')
            disconnect_system(devices, files_open=False)
            system_started = False
            pid = subprocess.run(['sudo', 'pgrep', 'python'], capture_output=True, text=True).stdout.split('\n')[:-1]
            for p in pid:
                subprocess.run(['kill', '-9', p])
    
        
        #*********** ****************************** ***********#
        if service != 'Mini':
            battery = {}
            for device in devices:
                state = device.state()
                if state['battery'] > 63:
                    battery_volts = 2 * ((state['battery']*3.3) / (2**10-1))
                else:
                    battery_volts = 2 * ((state['battery']*3.3) / (2**6-1))
                    
                battery[device.macAddress] = battery_volts
            battery_json = json.dumps(['BATTERY', battery])
            
            client.publish('rpi', battery_json)
            

        # Starting Acquisition LOOP =========================================================================
        try:
            while client.keepAlive == True:
                
                if pause_acq and not already_notified_pause:
                    
                    sync_param['mode'] = 0
                    
                    for i,device in enumerate(devices):
                        try:
                            device.stop()
                        except Exception as e:
                            print(e)
                            continue
                     
                    client.publish('rpi', "['PAUSED']")
                    already_notified_pause = True
                
                elif not pause_acq:
                    
                    if (already_notified_pause):
                        print('paused and restarting')
                        client.publish('rpi', "['RECONNECTING']")
                        
                    already_notified_pause = False
                    
                    if not system_started:
                    
                        try:
                        
                            _, t_disp, sync_param = start_system(devices, a_file, drift_log_file, opt['fs'], channels, sensors, save_fmt, header)
                            system_started = True
                            client.publish('rpi', "['ACQUISITION ON']")
        
                            t_display = process_data.decimate(t_disp, opt['fs'])
                            print('t_display: {}'.format(t_display))
                            json_data = json.dumps(['DATA', t_display, channels, sensors])
                            client.publish('rpi', json_data)
                            already_timed_out = False
                            
                            
                            
                        except Exception as e:
                            print(e)
                            pass
                    
                    # Acquisition LOOP =========================================================================
                    # try to read from the device--------------------------------------------------------------------------------------------------
                    
                    if write_annot:
                        print('SAVED ANNOT')
                        write_annot_file(annot_file, new_annot)
                        write_annot = False
                
                    try:
                        _, t_disp, a_file, drift_log_file, sync_param = run_system(devices, a_file, annot_file, drift_log_file, sync_param, directory, channels, sensors, opt['fs'], save_fmt, header)
                        
                        t_display = process_data.decimate(t_disp, opt['fs'])
                                    
                        json_data = json.dumps(['DATA', t_display, channels, sensors])
                        client.publish('rpi', json_data)
                        already_timed_out = False
                        
                    # Handle misconnection of the devices--------------------------------------------------------------------------------------------
                    except Exception as e:
                        print('')
                        print('')
                        print('The system has stopped running because ' + str(e) + '! Please check Modules!')
                        print('Trying to Reconnect....')
                        client.publish('rpi', "['RECONNECTING']")
                        
                        # Disconnect the system
                        disconnect_system(devices, a_file, annot_file, drift_log_file)
                        system_started = False
                        sync_param['mode'] = 0
                        
                        # Reconnect the devices
                        try:
                            #*********** Connection to BITalino devices ***********#
                            for i, mac in enumerate(opt['devices_mac']):
                            
                                init_connect_time = time.time()
                                print('Searching for Module...' + mac)
                                
                                while client.keepAlive == True:
    
                                    if (time.time() - init_connect_time) > 120:
                                        client.publish('rpi', "['STOPPED']")
                                        client.loop_stop()
                                        print('TIMEOUT')
                                        
                                        # Disconnect the system
                                        disconnect_system(devices, a_file, annot_file, drift_log_file)
                                        client.keepAlive == False
                                        pass
                                    
                                    try:
                                        if already_timed_out == True and time.time() - init_connect_time > 10:
                                            already_timed_out = False
                                        
                                        connected = False
                                        connected, devices = connect_device(mac, client, devices)

                                        if connected and mac in [d.macAddress for d in devices]:
                                            break
    
                                    except Exception as e:
                                        print(e)
                                        print('HERE Failed at connecting to BITalino')
                                        sync_param['mode'] = 0
                                        if already_timed_out == False and time.time() - init_connect_time > 10:
                                            timeout_json = json.dumps(['TIMEOUT', '{}'.format(mac)])
                                            client.publish('rpi', timeout_json)
                                            print('SENT TIMEOUT')
                                            already_timed_out = True
                                            init_connect_time = time.time()
                            
                            print('Devices in list: {}'.format([d.macAddress for d in devices]))
                            
                            a_file, annot_file, drift_log_file, save_fmt, header = open_file(directory, devices, channels, sensors, opt['fs'], saveRaw, service)
                        
                            # Acquisition LOOP =========================================================================
                                
                            for device in devices:
                                state = device.state()
                                if state['battery'] > 63:
                                    battery_volts = 2 * ((state['battery']*3.3) / (2**10-1))
                                else:
                                    battery_volts = 2 * ((state['battery']*3.3) / (2**6-1))
                                    
                                battery[device.macAddress] = battery_volts
                            battery_json = json.dumps(['BATTERY', battery])
                            
                            client.publish('rpi', battery_json)
                            
                            _, t_disp, sync_param = start_system(devices, a_file, drift_log_file, opt['fs'], channels, sensors, save_fmt, header)
                            system_started = True
                            print('The system is running again ...')
                            client.publish('rpi', "['ACQUISITION ON']")
                            
                            t_display = process_data.decimate(t_disp, opt['fs'])
                                        
                            json_data = json.dumps(['DATA', t_display, channels, sensors])
                            client.publish('rpi', json_data)
                            already_timed_out = False
                            
                        except:
                            pass
            
                else:
                    pass
            
        except KeyboardInterrupt:
            print('')
            print('You have stopped the acquistion. Saving all the files ...')
            client.publish('rpi', "['STOPPED']")
            client.loop_stop()
            
            # Disconnect the system
            disconnect_system(devices, a_file, annot_file, drift_log_file)
            client.keepAlive == False
            pass
    
            
            # -----------------------------------------------------------------------------------------------------------------------------------------
        
        print('')
        client.publish('rpi', "['STOPPED']")
        client.loop_stop()
        
        # Disconnect the system
        disconnect_system(devices, a_file, annot_file, drift_log_file)
        system_started = False
        print('You have stopped the acquistion. Saving all the files ...')
        time.sleep(3)
        pid = subprocess.run(['sudo', 'pgrep', 'python'], capture_output=True, text=True).stdout.split('\n')[:-1]
        for p in pid:
            subprocess.run(['kill', '-9', p])
        
    except Exception as e:
        print(e)
        client.publish('rpi', "['STOPPED']")
        client.loop_stop()
        
        # Disconnect the system
        disconnect_system(devices, a_file, annot_file, drift_log_file)
        system_started = False
        print('You have stopped the acquistion. Saving all the files ...')
        time.sleep(3)
        pid = subprocess.run(['sudo', 'pgrep', 'python'], capture_output=True, text=True).stdout.split('\n')[:-1]
        for p in pid:
            subprocess.run(['kill', '-9', p])
    
    # =========================================================================================================


class Frame:
    digital = [0] * 4
    seq = None
    a = [0] * 8

    def toMap(self):
        return {"sequence": self.seq, "analog": self.a, "digital": self.digital}

    def toString(self):
        return str(self.toMap())


# built-in
import time
from math import log2

# third-party
import bluetooth

# local
from . import frame
from . import state as st
from . import exceptions

TIMEOUT_IN_SECONDS = 3

# API_MODE
API_MODE_BITALINO = 1
API_MODE_SCIENTISST = 2
API_MODE_JSON = 3

# CHANNELS
AI1 = 1
AI2 = 2
AI3 = 3
AI4 = 4
AI5 = 5
AI6 = 6
AX1 = 7
AX2 = 8


def find():
    """
    Searches for Bluetooth devices in range.

    Parameters
    ----------
    void

    Returns
    -------
    devices : array
        List of found devices addresses
    """

    nearby_devices = bluetooth.discover_devices(lookup_names=True)

    c = 0
    devices = []
    for addr, name in nearby_devices:
        if "scientisst" in name.lower():
            devices += [addr]
            if c == 0:
                "Devices found:"
            c += 1
            print("{} - {}".format(addr, name))
    if c == 0:
        print("Found no devices")
    return devices


class ScientISST:
    """
    ScientISST Device class

    Parameters
    ----------
    address : String
        The device Bluetooth MAC address ("xx:xx:xx:xx:xx:xx")
    """

    __port = 1
    __sock = None
    __num_chs = 0
    __api_mode = 1
    __sample_rate = None
    __chs = [None] * 8
    __f = None

    def __init__(self, macAddress):
        self.macAddress = macAddress
        # Close socket if it exists
        if self.__sock:
            self.disconnect()

        print("Connecting to device...")
        # Create the client socket
        self.__sock = bluetooth.BluetoothSocket(bluetooth.RFCOMM)
        self.__sock.connect((self.macAddress, self.__port))
        print("Connected!")

    def version(self):
        """
        Gets the device firmware version string

        Parameters
        ----------
        void

        Returns
        -------
        version : string
            Firmware version
        """
        header = "ScientISST"
        headerLen = len(header)

        cmd = b"\x07"
        self.__send(cmd)
        version = ""
        while True:
            result = self.__recv(1)
            if result:
                if len(version) >= headerLen:
                    if result == b"\x00":
                        break
                    elif result != b"\n":
                        version += result.decode("utf-8")
                else:
                    result = result.decode("utf-8")
                    if result is header[len(version)]:
                        version += result
                    else:
                        version = ""
                        if result == header[0]:
                            version += result
            else:
                return

        print("ScientISST version: {}".format(version))
        return version

    def start(
        self, SamplingRate, analogChannels, file_name, simulated=False, api=API_MODE_SCIENTISST
    ):
        """
        Starts a signal acquisition from the device

        Parameters
        ----------
        sample_rate : int
            Sampling rate in Hz. Accepted values are 1, 10, 100 or 1000 Hz.
        channels : array
            Set of channels to acquire. Accepted channels are 1...6 for inputs A1...A6.
        file_name : string
            Name of the file where the live mode data will be written into.
        simulated : bool
            If true, start in simulated mode. Otherwise start in live mode. Default is to start in live mode.
        api : int
            The API mode, this API supports the ScientISST and JSON APIs.

        Returns
        -------
        void

        Exceptions
        ----------
        DeviceNotIdleError : if the device is already in acquisition mode.
        InvalidParameterError : if no valid API value is chosen or an incorrect array of channels is provided.
        """
        if self.__num_chs != 0:
            raise exceptions.DeviceNotIdleError()

        if api != API_MODE_SCIENTISST and api != API_MODE_JSON:
            raise exceptions.InvalidParameterError()

        self.__sample_rate = SamplingRate
        self.__num_chs = 0

        # Change API mode
        self.__changeAPI(api)

        # Sample rate
        sr = 0b01000011
        sr |= self.__sample_rate << 8
        self.__send(sr)

        if not analogChannels:  # channels is empty
            chMask = 0xFF  #  all 8 analog channels
            self.__num_chs = 8
        else:
            chMask = 0
            for ch in analogChannels:
                if ch <= 0 or ch > 8:
                    raise exceptions.InvalidParameterError()
                self.__chs[self.__num_chs] = ch  # Fill chs vector

                mask = 1 << (ch - 1)
                if chMask & mask:
                    raise exceptions.InvalidParameterError()

                chMask |= mask
                self.__num_chs += 1

        # Cleanup existing data in bluetooth socket
        self.__clear()

        if simulated:
            cmd = 0x02
        else:
            cmd = 0x01
        cmd |= chMask << 8

        self.__send(cmd)

        self.__packet_size = self.__getPacketSize()

        # Open file and write header
        self.__initFile(file_name)

    def read(self, num_frames):
        """
        Reads acquisition frames from the device.
        This method returns when all requested frames are received from the device, or when a timeout occurs.

        Parameters
        ----------
        num_frames : int
           Number of frames to retrieve from the device

        Returns
        -------
        frames : array
            List of Frame objects retrieved from the device

        Exceptions
        ----------
        DeviceNotInAcquisitionError : if the device is not in acquisition mode.
        NotSupportedError : if the device API is in BITALINO mode
        UnknownError : if the device stopped sending frames for some unknown reason.
        """

        frames = [None] * num_frames

        if self.__num_chs == 0:
            raise exceptions.DeviceNotInAcquisitionError()

        for it in range(num_frames):
            mid_frame_flag = 0
            bf = list(self.__recv(self.__packet_size))
            if not bf:
                raise exceptions.UnknownError("Esp stopped sending frames -> It stopped live mode on its own \n(probably because it can't handle this number of channels + sample rate)")

            #  if CRC check failed, try to resynchronize with the next valid frame
            while not self.__checkCRC4(bf, self.__packet_size):
                bf = bf[1:] + [None]
                #  checking with one new byte at a time
                result = self.__recv(1)
                bf[-1] = int.from_bytes(result, "big")

                if not bf[-1]:
                    return list(filter(lambda frame: frame, frames))   #  a timeout has occurred

            f = frame.Frame()
            frames[it] = f
            if self.__api_mode == API_MODE_SCIENTISST:
                # Get seq number and IO states
                f.seq = bf[-1] >> 4
                for i in range(4):
                    f.digital[i] = (bf[-2] & (0x80 >> i)) != 0

                # Get channel values
                byte_it=0
                for i in range(self.__num_chs):
                    curr_ch = self.__chs[self.__num_chs - 1 - i]

                    # If it's an AX channel
                    if curr_ch == AX1 or curr_ch == AX2:
                        f.a[curr_ch-1] = (
                            int.from_bytes(bf[byte_it : byte_it + 4], byteorder="little") & 0xFFFFFF
                        )
                        byte_it += 3

                    # If it's an AI channel
                    else:
                        if not mid_frame_flag:
                            f.a[curr_ch - 1] = (
                                int.from_bytes(bf[byte_it : byte_it + 2], byteorder="little")
                                & 0xFFF
                            )
                            byte_it += 1
                            mid_frame_flag = 1
                        else:
                            f.a[curr_ch - 1] = (
                                int.from_bytes(bf[byte_it : byte_it + 2], byteorder="little") >> 4
                            )
                            byte_it += 2
                            mid_frame_flag = 0
            elif self.__api_mode == API_MODE_JSON:
                print(bf)
            # d.Parse((const char*)buffer);

            # f.seq = 1;

            # for(int i = 0; i < num_chs; i++){
            # sprintf(memb_name, "AI%d", chs[i]);
            # f.a[i] = strtol(d[memb_name].GetString(), &junk, 10);
            # }

            # f.digital[0] = strtol(d["I1"].GetString(), &junk, 10);
            # f.digital[1] = strtol(d["I2"].GetString(), &junk, 10);
            # f.digital[2] = strtol(d["O1"].GetString(), &junk, 10);
            # f.digital[3] = strtol(d["O2"].GetString(), &junk, 10);
            # }
            else:
                raise exceptions.NotSupportedError()


            self.__writeFrameFile(f)

        return frames

    def stop(self):
        """
        Stops a signal acquisition.

        Parameters
        ----------
        void

        Returns
        -------
        void

        Exceptions
        ----------
        DeviceNotInAcquisitionError : if the device is not in acquisition mode.
        """
        if self.__num_chs == 0:
            raise exceptions.DeviceNotInAcquisitionError()

        cmd = b"\x00"
        self.__send(cmd)  # 0  0  0  0  0  0  0  0 - Go to idle mode

        self.__num_chs = 0
        self.__sample_rate = 0

        # Cleanup existing data in bluetooth socket
        self.__clear()

        # fclose(output_fd);
        self.__f.close()

    def battery(self, value=0):
        """
        Sets the battery voltage threshold for the low-battery LED.

        Parameters
        ----------
        value : int
            Battery voltage threshold. Default value is 0.
            Value | Voltage Threshold
            ----- | -----------------
                0 |   3.4 V
             ...  |   ...
               63 |   3.8 V

        Returns
        -------
        void

        Exceptions
        ----------
        DeviceNotIdleError : if the device is in acquisition mode.
        InvalidParameterError : if an invalid battery threshold value is given.
        """
        if self.__num_chs != 0:
            raise exceptions.DeviceNotIdleError()

        if value < 0 or value > 63:
            raise exceptions.InvalidParameterError()

        cmd = value << 2
        # <bat threshold> 0 0 - Set battery threshold
        self.__send(cmd)

    def trigger(self, digital_output):
        """
        Assigns the digital outputs states.

        Parameters
        ----------
        digital_output : array
            Vector of booleans to assign to digital outputs, starting at first output (O1).

        Returns
        -------
        void

        Exceptions
        ----------
        InvalidParameterError : if the length of the digital_output array is different from 2.
        """
        length = len(digital_output)

        if length != 2:
            raise exceptions.InvalidParameterError()

        cmd = 0xB3  # 1  0  1  1  O2 O1 1  1 - Set digital outputs

        for i in range(length):
            if digital_output[i]:
                cmd |= 0b100 << i

        self.__send(cmd)

    def dac(self, pwm_output):
        """
        Assigns the analog (PWM) output value (%ScientISST 2 only).

        Parameters
        ----------
        pwm_output : int
            Analog output value to set (0...255).

        Returns
        -------
        void

        Exceptions
        ----------
        InvalidParameterError : if the pwm_output value is outside of its range, 0-255.
        """
        if pwm_output < 0 or pwm_output > 255:
            raise exceptions.InvalidParameterError()

        cmd = 0xA3  # 1  0  1  0  0  0  1  1 - Set dac output

        cmd |= pwm_output << 8
        self.__send(cmd)

    # TODO: test with ScientISST Sense v2
    def state(self):
        """
        Returns current device state (%ScientISST 2 only).

        Parameters
        ----------
        void

        Returns
        -------
        state : State
            Current device state

        Exceptions
        ----------
        DeviceNotIdleError : if the device is in acquisition mode.
        ContactingDeviceError : if there is an error contacting the device.
        """
        if self.__num_chs != 0:
            raise exceptions.DeviceNotIdleError()

        cmd = 0x0B
        self.__send(cmd)
        # 0  0  0  0  1  0  1  1 - Send device status

        # if (recv(&statex, sizeof statex) != sizeof statex)    # a timeout has occurred
        # throw Exception(Exception::CONTACTING_DEVICE);
        result = self.__recv(16)
        if not result or not self.__checkCRC4(result, 16):
            raise exceptions.ContactingDeviceError()

        state = st.State()
        print(result)

        # for(int i = 0; i < 6; i++)
        # state.analog[i] = statex.analog[i];

        # state.battery = statex.battery;
        # state.batThreshold = statex.batThreshold;

        # for(int i = 0; i < 4; i++)
        # state.digital[i] = ((statex.portsCRC & (0x80 >> i)) != 0);

        # return state;

    def disconnect(self):
        """
        Disconnects from a ScientISST device. If an aquisition is running, it is stopped

        Parameters
        ----------
        void

        Returns
        -------
        void
        """
        if self.__num_chs != 0:
            self.stop()
        self.__sock.close()
        self.__sock = None
        print("Disconnected")


    def __getPacketSize(self):
        packet_size = 0

        if self.__api_mode == API_MODE_SCIENTISST:
            num_intern_active_chs = 0
            num_extern_active_chs = 0

            for ch in self.__chs:
                if ch:
                    # Add 24bit channel's contributuion to packet size
                    if ch == AX1 or ch == AX2:
                        num_extern_active_chs += 1
                    # Count 12bit channels
                    else:
                        num_intern_active_chs += 1

            # Add 24bit channel's contributuion to packet size
            packet_size = 3 * num_extern_active_chs

            # Add 12bit channel's contributuion to packet size
            if not (num_intern_active_chs % 2):  # If it's an even number
                packet_size += (num_intern_active_chs * 12) / 8
            else:
                packet_size += (
                    (num_intern_active_chs * 12) - 4
                ) / 8  # -4 because 4 bits can go in the I/0 byte
            # for the I/Os and seq+crc bytes
            packet_size += 2

        elif self.__api_mode == API_MODE_JSON:
            for i in range(self.__num_chs):
                # If it's internal ch
                if self.__chs[i]<=6:
                    # sprintf(aux_str, "AI%d", chs[i]);
                    # member_name.SetString(aux_str, d.GetAllocator());
                    # member_value.SetString(value_internal_str, d.GetAllocator());
                    # d.AddMember(member_name, member_value, d.GetAllocator());
                    packet_size+=3  # AI%d
                    packet_size+=2  # 0-4095 = 12 bits <= 2 bytes
                else:
                    packet_size+=3  # AX%d
                    packet_size+=4  # 0-16777215 = 24 bits <= 4 bytes
            # Add IO state json objects
            # d.AddMember("I1", "0", d.GetAllocator());
            # d.AddMember("I2", "0", d.GetAllocator());
            # d.AddMember("O1", "0", d.GetAllocator());
            # d.AddMember("O2", "0", d.GetAllocator());
            packet_size += 3    # I1 + 0|1
            packet_size += 3    # I2 + 0|1
            packet_size += 3    # O1 + 0|1
            packet_size += 3    # O1 + 0|1

        else:
            raise exceptions.NotSupportedError()

        return int(packet_size)

    def __initFile(self, filename):
        self.__f = open(filename, "w")
        self.__f.write("NSeq, I1, I2, O1, O2, ")
        for i in range(self.__num_chs):
            ch = int(self.__chs[i])
            if ch == AX1 or ch == AX2:
                if i == self.__num_chs - 1:
                    self.__f.write("AX{}".format(ch))
                else:
                    self.__f.write("AX{}, ".format(ch))
            else:
                if i == self.__num_chs - 1:
                    self.__f.write("AI{}".format(ch))
                else:
                    self.__f.write("AI{}, ".format(ch))
        self.__f.write("\n")

    def __writeFrameFile(self, f):
        self.__f.write(
            "{}, {}, {}, {}, {}, ".format(
                f.seq, f.digital[0], f.digital[1], f.digital[2], f.digital[3]
            )
        )

        for i in range(self.__num_chs):
            if i == self.__num_chs - 1:
                # self.__f.write("%d", esp_adc_cal_raw_to_voltage(f.a[chs[i]-1], &adc1_chars))
                self.__f.write("{}".format(f.a[self.__chs[i] - 1]))
            else:
                # self.__f.write("%d, ", esp_adc_cal_raw_to_voltage(f.a[chs[i]-1], &adc1_chars))
                self.__f.write("{}, ".format(f.a[self.__chs[i] - 1]))
        self.__f.write("\n")

    def __changeAPI(self, api):
        if self.__num_chs and self.__num_chs != 0:
            raise exceptions.DeviceNotIdleError()

        if api <= 0 or api > 3:
            raise exceptions.InvalidParameterError()

        self.__api_mode = api

        api <<= 4
        api |= 0b11

        self.__send(api)

    def __checkCRC4(self, data, length):
        CRC4tab = [0, 3, 6, 5, 12, 15, 10, 9, 11, 8, 13, 14, 7, 4, 1, 2]
        crc = 0
        for i in range(length - 1):
            b = data[i]
            crc = CRC4tab[crc] ^ (b >> 4)
            crc = CRC4tab[crc] ^ (b & 0x0F)

        # CRC for last byte
        crc = CRC4tab[crc] ^ (data[-1] >> 4)
        crc = CRC4tab[crc]

        return crc == (data[ - 1] & 0x0F)

    def __send(self, command):
        """
        Send data
        """
        if type(command) is int:
            if command != 0:
                command = command.to_bytes(
                    int(log2(command) // 8 + 1), byteorder="little"
                )
            else:
                command = b"\x00"
        if self.__sock:
            time.sleep(0.150)
            # print bytes sent
            # print("{} bytes sent: ".format(len(command))+ " ".join("{:02x}".format(c) for c in command))
            self.__sock.send(command)
        else:
            raise exceptions.ContactingDeviceError()

    def __recv(self, nrOfBytes):
        """
        Receive data
        """
        result = None
        self.__sock.settimeout(TIMEOUT_IN_SECONDS)
        try:
            result = self.__sock.recv(nrOfBytes)
            pass
        except bluetooth.btcommon.BluetoothError:
            pass
        self.__sock.settimeout(None)
        return result

    def __clear(self):
        """
        Clear the device buffer
        """
        while self.__recv(1):
            pass


# built-in
import time
import datetime

# third-party
import numpy as np

# local
from epibox.common.write_file import write_file
from epibox.common.read_modules import read_modules


def start_system(devices, a_file, drift_log_file, fs, mac_channels, sensors, save_fmt, header):

    dig_Out = 0
    now = datetime.now()
    sync_param = {'flag_sync' : 0 , 'inittime' : time.time(), 'strtime': time.time(), 'sync_time' : now.strftime("%Y-%m-%d %H:%M:%S.%f").rstrip('0'), 'dig_Out' : dig_Out, 'close_file' : 0, 'mode': 0, 'diff': 1000, 'save_log': 1, 'count_a': 1000, 'sync_append': 0}
    # mode: 0 if not started acquisition yet (or if paused) and 1 otherwise (used to write in drift_log_file)
    
    for i in range(len(devices)):
        sync_param['sync_arr_'+chr(ord('@')+i+1)] = np.zeros(1000, dtype = float)


    # Initialize devices
    for device in devices:
        if header['service'] == 'Bitalino' or header['service'] == 'Mini':
            channels = [int(elem[1])-1 for elem in mac_channels if elem[0]==device.macAddress]
        else:
            channels = [int(elem[1]) for elem in mac_channels if elem[0]==device.macAddress]
                
        print('channels: {}'.format(channels))
        device.start(SamplingRate=fs, analogChannels=channels)
        print('START SYSTEM')

  
    now = datetime.now()
    print('start {}'.format(datetime.now()))
    sync_param['sync_time'] = now.strftime("%Y-%m-%d %H:%M:%S.%f")
    print(now.strftime("%Y-%m-%d %H:%M:%S.%f"))
    t, t_str, t_display = read_modules(devices, mac_channels, sensors, header)
 
    sync_param['save_log'] = 1
            
    write_file(t, a_file, drift_log_file, sync_param, str(sync_param['inittime']), save_fmt)
    print('System initiated')                    
    
    return t, t_display, sync_param
    



def close_file(a_file, annot_file, log_drift_file):
    close_acq_file(a_file)
    close_drift_log_file(log_drift_file)

def close_acq_file(a_file):
	a_file.close()


def close_drift_log_file(drift_log_file):
	drift_log_file.close()




# built-in
import ast
import string 
import random
import os
import subprocess
import json
import shutil

# third-party
import paho.mqtt.client as mqtt

# local
from epibox.common.connect_device import connect_device
from epibox.run import run_bitalino

def random_str(length):

    letters = string.ascii_letters

    return ''.join(random.choice(letters) for i in range(length))

def on_message(client, userdata, message):

    global sys_args
    global devices

    message = str(message.payload.decode("utf-8"))
    print(message)
    message = ast.literal_eval(message)

    if message == ['Send default']:

        ######## Default MAC addresses ########

        with open('/home/pi/Documents/epibox/listMAC.json', 'r') as json_file:
            listMAC = json_file.read()

        listMAC = ast.literal_eval(listMAC)
        listMAC2 = "'DEFAULT MAC','{}','{}'".format(list(listMAC.values())[0], list(listMAC.values())[1])

        client.publish(topic='rpi', qos=2, payload=listMAC2)

        ######## Available drives ########

        listDrives = ['DRIVES']
        drives = os.listdir('/media/pi')

        for drive in drives:
            total, _ , free = shutil.disk_usage('/media/pi/{}'.format(drive))
            listDrives += ['{} ({:.1f}% livre)'.format(drive, (free/total)*100)]

        total, _ , free = shutil.disk_usage('/')
        listDrives += ['RPi ({:.1f}% livre)'.format((free/total)*100)]

        client.publish(topic='rpi', qos=2, payload="{}".format(listDrives))

        
        ######## Default configurations ########

        with open('/home/pi/Documents/epibox/config_default.json', 'r') as json_file:
            defaults = json_file.read()

        defaults = ast.literal_eval(defaults)
        config = json.dumps(['DEFAULT CONFIG', defaults])

        client.publish(topic='rpi', qos=2, payload=config)


    ######## New default configuration ########

    elif message[0] == 'NEW CONFIG DEFAULT':
        config = message[1]
        defaults = {'initial_dir': config[0], 'fs': config[1], 'channels': config[2], 'saveRaw': config[3]}

        with open('/home/pi/Documents/epibox/config_default.json', 'w') as json_file:
            json.dump(defaults, json_file)

            

    elif message[0] == 'NEW MAC':
        listMAC = message[1]

        with open('/home/pi/Documents/epibox/listMAC.json', 'w') as json_file:
            json.dump(listMAC, json_file)


    ######## Update time ########

    elif message[0] == 'TIME':
        subprocess.run(["sudo", "date", "-s", message[1]])
        subprocess.run(["sudo", "date"], capture_output=True, text=True)


    ##### Connect devices #####

    elif message[0] == 'CONNECT':
        mac = message[1]

        if mac != ' ' and mac != '' and message[1] not in sys_args['devices_mac']:
            sys_args['devices_mac'] += [mac]

        _, devices = connect_device(mac, client, devices)


    ####### Set configurations ########

    elif message[0] == 'FOLDER':

        if message[1] == 'RPi':
            folder = '/home/pi/Documents/epibox/acquisitions'

        else: 
            folder = '/media/pi/' + message[1] + '/acquisitions'

        sys_args['initial_dir'] = folder

    
    elif message[0] == 'FS':
        fs = message[1]
        sys_args['fs'] = fs

    
    elif message[0] == 'SAVE RAW':
        saveRaw = message[1]
        sys_args['saveRaw'] = saveRaw


    elif message[0] == 'EPI SERVICE':
        epi_service = message[1]
        sys_args['service'] = epi_service


    elif message[0] == 'CHANNELS':
        channels = message[1]
        sys_args['channels'] = channels

        client.publish(topic='rpi', qos=2, payload=str(['RECEIVED CONFIG']))

    
    elif message[0] == 'ID':
        patient_id = message[1]
        sys_args['patient_id'] = patient_id


    ####### System actions #######

    elif message[0] == 'TURN OFF':
        print('TURNING OFF RPI')
        client.publish(topic='rpi', qos=2, payload=str(['TURNED OFF']))

    elif message[0] == 'TURNED OFF':
        subprocess.run(['sudo', 'shutdown', '-h', 'now'])


    elif message[0] == 'START':
        client.keepAlive = False

 
def main():

    global devices
    devices = []

    global sys_args
    sys_args = {'initial_dir': None, 'fs': None, 'channels': None, 'saveRaw': None, 'devices_mac': [], 'patient_id': None, 'service': None}

    client_name = random_str(6)
    print('Client name (startup):', client_name)
    host_name = '192.168.0.10'
    topic = 'rpi'

    client = mqtt.Client(client_name)
    setattr(client, 'keepAlive', True)

    client.username_pw_set(username='preepiseizures', password='preepiseizures')
    client.connect(host_name)
    client.subscribe(topic, 1)
    client.on_message = on_message
    client.loop_start()

    print('Successfully subcribed to topic', topic)

    while client.keepAlive == True:
        continue

    else:
        client.loop_stop()

        with open('/home/pi/Documents/epibox/args.json', 'w') as json_file:
            json.dump(sys_args, json_file)

        run_bitalino.main(devices)

    
if __name__ == '__main__':

    main()





def header2scientisst(filename, file_time, file_date, devices, mac_channels, sensors, fs, saveRaw):
    
    filename.write('# OpenSignals Text File Format' + '\n')
    
    mac_dict = {}
    resolution = {}
    fmt = []# get format to save values in txt
    
    for n_device,device in enumerate(devices):
        
        fmt += ['%i']
        mac_dict[device.macAddress] = {}
        
        mac_dict[device.macAddress]['sensor'] = []
        for i,elem in enumerate(mac_channels):
            if elem[0]==device.macAddress:
                mac_dict[device.macAddress]['sensor'] += [sensors[i]]
                if saveRaw:
                    fmt += ['%i']
                else:
                    fmt += ['%.2f']
        
        mac_dict[device.macAddress]['device name'] = 'Device '+str(n_device+1) 
        
        aux = ['A'+elem[1] for elem in mac_channels if elem[0]==device.macAddress]
        mac_dict[device.macAddress]['column'] = ["nSeq"]+aux
        
        mac_dict[device.macAddress]['sync interval'] = 2 #???
        
        mac_dict[device.macAddress]['start time'] = file_time
        
        mac_dict[device.macAddress]['device connection'] = device.macAddress
                
        mac_dict[device.macAddress]['channels'] = [int(elem[1]) for elem in mac_channels if elem[0]==device.macAddress]
        
        mac_dict[device.macAddress]['date'] = file_date
        
        mac_dict[device.macAddress]['firmware version'] = device.version() 
        
        mac_dict[device.macAddress]['device'] = 'scientisst_sense'
        
        mac_dict[device.macAddress]['sampling rate'] = fs
        
        if saveRaw:
            mac_dict[device.macAddress]['label'] = ['RAW' for i,elem in enumerate(mac_channels) if elem[0]==device.macAddress]
        else:
            mac_dict[device.macAddress]['label'] = [sensors[i] for i,elem in enumerate(mac_channels) if elem[0]==device.macAddress]
        
        
        mac_dict[device.macAddress]['resolution'] = [4] + [12]*len(mac_channels)
        resolution[device.macAddress] = mac_dict[device.macAddress]['resolution']

    header = {'resolution': resolution, 'saveRaw': saveRaw, 'service': 'Sense'}
    
    print("# " + str(mac_dict) + '\n')
    filename.write("# " + str(mac_dict) + '\n')
    filename.write('# EndOfHeader' + '\n')
    
    return tuple(fmt), header
